%% $Id: BagOps.oz,v 1.3 2010/02/15 04:18:59 leavens Exp leavens $
declare
%%-------------------------------------------------------
%% Start of provided functions section.
%% Do not change any of the following provided functions.
%% The code you are to write follows this section.
%%-------------------------------------------------------

%% We are representing bags as association lists with
%% pairs of the form <Element>#<Count>,
%% without duplicates among the elements
%% and with all counts strictly greater than 0.
%% Elements are the same if they are == to each other.
%% <Bag T> ::= nil
%%         |   T#<Int> '|' <Bag T>

%% Return the empty bag
fun {EmptyBag} nil end

%% Return a set containing exactly the elements of L
%% The multiplicity of each element is the number of times it occurs in the list
fun {AsBag L}
   case L of
      E|Es then {Add {AsBag Es} E 1}
   else nil
   end
end

%% Return a new bag containing all the elements of Bag
%% and NumTimes more copies of Elem
fun {Add Bag Elem NumTimes}
   case Bag of
      (E#Count)|Es then if E == Elem
                        then (E#Count+NumTimes)|Es
                        else (E#Count)|{Add Es Elem NumTimes}
                        end
   else (Elem#NumTimes)|nil
   end
end

%% Is this Bag empty?
fun {IsEmpty Bag}
   case Bag of
      _|_ then false
   else true
   end
end

%% How many times does Elem occur in Bag?
fun {Multiplicity Bag Elem}
   case Bag of
      (E#Count)|Es then
      if Elem == E then Count else {Multiplicity Es Elem} end
   else 0
   end
end

%% Is B1 a (not necessarily proper) subbag of B2?
%% That is, are all elements of B1 also in B2?
%% Put another way, for each element of B1,
%% is its multiplicity in B1 no larger than its multiplicity in B2?
fun {IsSubbag B1 B2}
   case B1 of
      (E#Count)|Es then Count =< {Multiplicity B2 E} andthen {IsSubbag Es B2}
   else true
   end
end

%% Are bags B1 and B2 equal?
fun {Equal B1 B2}
   {IsSubbag B1 B2} andthen {IsSubbag B2 B1}
end

%%----------------------------------
%% End of provided code.
%%----------------------------------

%%---------------------------------
%% Fill in your code for each of the functions below.
%%---------------------------------

%% Return the number of elements in this bag.
%% That is, return the sum of all the multiplicities of elements in this bag.
fun {Size L}
   raise notImplementedYet end %% <-- Replace with your code
end

%% Return a new bag that contains all the elements of B,
%% except NumTimes fewer of Elem
fun {Remove B Elem NumTimes}
   raise notImplementedYet end %% <-- Replace with your code
end

%% Return a bag that contains just the elements of both B1 and B2
%% That is, the multiplicity of an element in the result is the sum of
%% the multiplicities of that element in B1 and B2.
%% (Hint: If you recur on B1, then you can use B2 as an accumulator.)
fun {Union B1 B2}
   raise notImplementedYet end %% <-- Replace with your code
end

%% Return a bag that contains just the elements of B1 that are not in B2
%% That is, the multiplicity of an element in the result is the multiplicity
%% of that element in B1 minus its multiplicity in B2.
fun {Minus B1 B2}
   raise notImplementedYet end %% <-- Replace with your code
end

%% Return a bag that contains just the elements of B1 that are also in B2
%% That is, the multiplicity of an element in the result is the
%% minimum of the multiplicity of that element in B1 and B2.
fun {Intersect B1 B2}
   raise notImplementedYet end %% <-- Replace with your code
end

%% Return a bag that contains just the elements of all the bags in Bags
%% That is, the multiplicity of an element in the result is
%% the sum of its multiplicities in any of the bags in Bags.
fun {UnionOfAll Bags}
   raise notImplementedYet end %% <-- Replace with your code
end
